//   ===== Phase 1 code =====
function DHKE_requestPhase1Handle(%prime)
{
	%handle = -1+$DHKE_WorkerCount++;
	if(%prime $= "") $DHKE_PrimeHandle[%handle] = DHKE_requestPrime();
	else
	{
		$DHKE_PrimeValue[%handle] = %prime | 0;
		$DHKE_FinalValue[%handle] = 1;
	}
	$DHKE_SecretValue[%handle] = getRandom(65536, 262144);
	$DHKE_WorkSchedule[%handle] = schedule(50, 0, "DHKE_phase1Worker", %handle);
	return %handle;
}

function DHKE_phase1Worker(%handle)
{
	cancel($DHKE_WorkSchedule[%handle]);
	if($DHKE_PrimeValue[%handle] $= "")
	{
		if(DHKE_primeReady($DHKE_PrimeHandle[%handle]))
			$DHKE_PrimeValue[%handle] = DHKE_getPrime($DHKE_PrimeHandle[%handle]);
		$DHKE_FinalValue[%handle] = 1;
	}
	else
	{
		for(%i=0;%i<1024;%i++)
		{
			if($DHKE_Iterations[%handle] < $DHKE_SecretValue[%handle])
			{
				%in = $DHKE_FinalValue[%handle];
				%final = ((%in | 0) * 5) % $DHKE_PrimeValue[%handle];
				$DHKE_FinalValue[%handle] = %final | 0;
				$DHKE_Iterations[%handle]++;
			}
			else
			{
				$DHKE_IsReady1_[%handle] = 1;
				return;
			}
		}
	}
	$DHKE_WorkSchedule[%handle] = schedule(50, 0, "DHKE_phase1Worker", %handle);
}

function DHKE_phase1Ready(%handle)
{ return !!$DHKE_IsReady1_[%handle]; }

function DHKE_getPhase1(%handle)
{
	if($DHKE_IsReady1_[%handle])
		return $DHKE_FinalValue[%handle];
	else return -1;
}

//   ===== Phase 2 code =====
function DHKE_requestPhase2Handle(%shared, %handle)
{
	if(!$DHKE_IsReady1_[%handle])
		return 0;
	$DHKE_SharedValue[%handle] = %shared | 0;
	$DHKE_WorkSchedule[%handle] = schedule(50, 0, "DHKE_phase2Worker", %handle);
	$DHKE_SharedSecret[%handle] = 1;
	$DHKE_Iterations[%handle] = 0;
	return 1;
}

function DHKE_phase2Worker(%handle)
{
	cancel($DHKE_WorkSchedule[%handle]);
	for(%i=0;%i<128;%i++)
	{
		if($DHKE_Iterations[%handle] < $DHKE_SecretValue[%handle])
		{
			%in = $DHKE_SharedSecret[%handle];
			%final = DHKE_phase2Roll(%in | 0, $DHKE_SharedValue[%handle] | 0, $DHKE_PrimeValue[%handle] | 0);
			//%final = ((%in | 0) * 5) % $DHKE_PrimeValue[%handle];
			$DHKE_SharedSecret[%handle] = %final | 0;
			$DHKE_Iterations[%handle]++;
		}
		else
		{
			$DHKE_IsReady2_[%handle] = 1;
			return;
		}
	}
	$DHKE_WorkSchedule[%handle] = schedule(50, 0, "DHKE_phase2Worker", %handle);
}

function DHKE_phase2Roll(%in, %mult, %mod)
{
	%mod |= 0;
	if(%mod <= 0)
		return -1;
	%in |= 0;
	%mult |= 0;
	
	//Break our inputs into their component chunks
	%chunk0 = %in & 65535;
	%chunk1 = (%in >> 16) & 65535;
	%chunk2 = %mult & 65535;
	%chunk3 = (%mult >> 16) & 65535;
	%chunk4 = %mod & 65535;
	%chunk5 = (%mod >> 16) & 65535;
	
	//Multiply %in by %mult to get four chunks
	%aChunk0 = (%chunk0 * %chunk2) | 0;
	%aChunk1 = (%chunk0 * %chunk3) | 0;
	%aChunk2 = (%chunk1 * %chunk2) | 0;
	%aChunk3 = (%chunk1 * %chunk3) | 0;
	
	//Interpolate the result chunks to get a 64-bit number
	%bChunk0 = %aChunk0 & 65535;
	%bChunk1 = (%aChunk0 >> 16) & 65535;
	%bChunk1 += %aChunk1 & 65535;
	%bChunk1 += %aChunk2 & 65535;
	%bChunk2 = (%bChunk1 >> 16) & 65535;
	%bChunk1 &= 65535;
	%bChunk2 += (%aChunk1 >> 16) & 65535;
	%bChunk2 += (%aChunk2 >> 16) & 65535;
	%bChunk2 += %aChunk3 & 65535;
	%bChunk3 = (%bChunk2 >> 16) & 65535;
	%bChunk3 += (%aChunk3 >> 16) & 65535;
	%bChunk2 &= 65535;
	
	//echo(%aChunk3 SPC %aChunk2 SPC %aChunk1 SPC %aChunk0);
	//echo(%bChunk3 SPC %bChunk2 SPC %bChunk1 SPC %bChunk0);
	
	//Shifts the modulo value over to the left
	for(%i=0;%i<32;%i++)
	{
		if(%mod & 0x40000000)
			break;
		%mod <<= 1;
	}
	if(!(%mod & 0x40000000))
		return -1;
	
	//Calculates the modulo remainder
	%cChunk3 = (%mod >> 16) & 65535;
	%cChunk2 = %mod & 65535;
	%cChunk1 = 0;
	%cChunk0 = 0;
	for(%i=0;%i<64;%i++)
	{
		if(%bChunk3 > %cChunk3)
		{
			%bChunk3 -= %cChunk3;
			%bChunk2 -= %cChunk2;
			%bChunk1 -= %cChunk1;
			%bChunk0 -= %cChunk0;
			if(%bChunk2 < 0)
			{
				%bChunk2 += 65536;
				%bChunk3--;
			}
			if(%bChunk1 < 0)
			{
				%bChunk1 += 65536;
				%bChunk2--;
			}
			if(%bChunk0 < 0)
			{
				%bChunk0 += 65536;
				%bChunk1--;
			}
		}
		else if(%bChunk3 == %cChunk3)
		{
			if(%bChunk2 > %cChunk2)
			{
				%bChunk3 -= %cChunk3;
				%bChunk2 -= %cChunk2;
				%bChunk1 -= %cChunk1;
				%bChunk0 -= %cChunk0;
				if(%bChunk1 < 0)
				{
					%bChunk1 += 65536;
					%bChunk2--;
				}
				if(%bChunk0 < 0)
				{
					%bChunk0 += 65536;
					%bChunk1--;
				}
			}
			else if(%bChunk2 == %cChunk2)
			{
				if(%bChunk1 > %cChunk1)
				{
					%bChunk3 -= %cChunk3;
					%bChunk2 -= %cChunk2;
					%bChunk1 -= %cChunk1;
					%bChunk0 -= %cChunk0;
					if(%bChunk0 < 0)
					{
						%bChunk0 += 65536;
						%bChunk1--;
					}
				}
				else if(%bChunk1 == %cChunk1 && %bChunk0 >= %cChunk0)
				{
					%bChunk3 -= %cChunk3;
					%bChunk2 -= %cChunk2;
					%bChunk1 -= %cChunk1;
					%bChunk0 -= %cChunk0;
				}
			}
		}
		
		//Shifts the whole array one bit to the right
		if(%cChunk0 & 1)
			break;
		%cChunk0 >>= 1;
		if(%cChunk1 & 1)
			%cChunk0 |= 32768;
		%cChunk1 >>= 1;
		if(%cChunk2 & 1)
			%cChunk1 |= 32768;
		%cChunk2 >>= 1;
		if(%cChunk3 & 1)
			%cChunk2 |= 32768;
		%cChunk3 >>= 1;
	}
	
	return %bChunk0 | (%bChunk1 << 16);
}

function DHKE_phase2Ready(%handle)
{ return !!$DHKE_IsReady2_[%handle]; }

function DHKE_getPhase2(%handle)
{
	if($DHKE_IsReady2_[%handle])
		return $DHKE_SharedSecret[%handle];
	else return -1;
}

//   ===== Prime number generation =====
function DHKE_RandomPrimeGenLoop()
{
	cancel($DHKE_RandomPrimeGenLoop);
	if($DHKE_GenPrimes < $DHKE_NeedPrimes + 1024)
		for(%i=0;%i<128;%i++)
		{
			%base = getRandom(1, 999999);
			%base = getSubStr("000000", 0, 6 - strLen(%base)) @ %base;
			%num = (getRandom(17, 425) @ %base) | 1;
			if(DHKE_isPrime(%num))// && DHKE_isPrime(2 * %num + 1))
			{
				$DHKE_GenPrime[-1+$DHKE_GenPrimes++] = %num | 0; //(2 * %num + 1) | 0;
				break;
			}
		}
	$DHKE_RandomPrimeGenLoop = schedule(16, 0, "DHKE_RandomPrimeGenLoop");
}

function DHKE_requestPrime()
{ return -1+$DHKE_NeedPrimes++; }

function DHKE_primeReady(%handle)
{ return $DHKE_GenPrimes > %handle; }

function DHKE_getPrime(%handle)
{
	if($DHKE_GenPrimes > %handle)
		return $DHKE_GenPrime[%handle];
	else return -1;
}

function DHKE_isPrime(%num)
{
	if(!$DHKE_MadePrimeTable)
		DHKE_GenPrimeTable();
	%num |= 0;   //Forces Torque to recast this as an Int32 value
	for(%i=0;%i<$DHKE_Primes;%i++)
		if(!(%num%$DHKE_Prime[%i]))
			return false;
	return true;
}

exec("./DHKE_PrimeTable.cs");
if(!isEventPending($DHKE_RandomPrimeGenLoop))
	DHKE_RandomPrimeGenLoop();